(*
https://www.smlnj.org/doc/smlnj-lib/Util/sig-ORD_MAP.html#str:AtomMap
https://smlfamily.github.io/Basis/list.html#SIG:LIST.exists:VAL
https://www.smlnj.org/doc/smlnj-lib/Util/str-Atom.html
https://smlfamily.github.io/Basis/option.html
*)

datatype term = Var of Atom.atom  (* Variables *)
              | Fn of term list   (* Functional symbols *)
               
fun arity (t:term) : int option =
  case t of
    Fn xs => SOME (List.length xs)
  | Var _ => NONE

(*
We can find the arity of a functional symbol by finding
the length of its list argument.
 *)

(* Type of equations *)
type eqn = term * term

fun zip (l1:'a list) (l2:'a list) : ('a*'a) list =
  case (l1, l2) of
      (x::xs, y::ys) => (x,y) :: (zip xs ys)
  | (_, _)  => []
(*
- zip [1,2] [3,4];
val it = [(1,3),(2,4)] : (int * int) list
*)
            
(*
 - r: replacement term
 - v: variable whose every occurrance in [t] needs to be replaced
      with [r]
 - t: term in which substitution is to be done.
 *)
fun subst (r:term) (v:Atom.atom) (t:term) : term =
  case t of
    Fn xs =>
    Fn (List.map
           (fn elem:term => subst r v elem) xs)
  | Var x => if Atom.same (x, v) then r
             else Var x                       

fun updateUnifier (m:term AtomMap.map) (v:Atom.atom) (t:term)
  : (term AtomMap.map) option =
  let val m' = AtomMap.insert (m, v, t) in
    SOME (AtomMap.map (fn elem:term => subst t v elem) m)
  end  
            
(* Check if a variable [v] appears in a term [t] *)
fun varInTerm (v:Atom.atom) (t:term) : bool =
  case t of
  Var x => if Atom.same (x, v) then true else false
  | Fn xs => List.exists (fn elem:term => varInTerm v elem) xs
             
(* Attempt to unify a term [t] and a variable [v],
   given a set of already known substitutions [s] *)
fun unifyVar (v:Atom.atom) (t:term)
    (m:term AtomMap.map) : (term AtomMap.map) option =
  if varInTerm v t then NONE
  else updateUnifier m v t
	  (*
  case t of
    (* TODO: What to do when both are variables? *)
    Var x => if Atom.same (x, v) then updateUnifier m v t
             else NONE
  | Fn _ => if varInTerm updateUnifier m v t
  | Fn xs => NONE
      *)

(* Unify an equation *)
fun unify (q:eqn) (m:term AtomMap.map)
  : (term AtomMap.map) option =
  case q of
    (Fn xs, Fn ys) =>
      List.foldl
          (fn (elem:term*term, acc:(term AtomMap.map) option) =>
            if Option.isSome acc then
              unify elem (Option.valOf acc)
            else NONE)
      (SOME m) (zip xs ys)
  | (Var x, t) => unifyVar x t m
  | (t, Var x) => unifyVar x t m
(*
  | (Fn [], Fn []) => SOME m 
    (* Constants are okay *)

  | (Fn _, Fn _) => NONE
    (* We deal only with first order unification *)
*)

						   
(*
Control.Print.printDepth := 100;
Control.Print.printLength := 100;
Control.Print.stringDepth := 200;
*)

val _ = Control.Print.printDepth := 100;
val _ = Control.Print.printLength := 100;
val _ = Control.Print.stringDepth := 200;

val a0 = Fn [Fn [], Var (Atom.atom "x")]
val b0 = Fn [Var (Atom.atom "x"), Fn []]
val u0 = unify (a0, b0) AtomMap.empty


val a1 = Fn [Fn [], Var (Atom.atom "x")]
val b1 = Fn [Var (Atom.atom "x"), Fn []]
val u1 = unify (a1, b1) AtomMap.empty
val u1' = AtomMap.listItemsi (Option.valOf u1);

